
import time
import numpy as np
from srsgui import Task
from srsgui import InstrumentInput, FloatListInput
from srsgui import TimePlot

from srsinst.qcm import get_qcm


class QCMMonitorTask(Task):
    """Monitor QCM200 frequency and resistance 
    """
    
    InstrumentName = 'qcm to monitor'
    GateTime = 'gate time (s)'
    GateTimeList = [0.1, 1.0, 10.0]

    input_parameters = {
        InstrumentName: InstrumentInput(),
        GateTime: FloatListInput(GateTimeList, '{:.1f}')
    }

    def setup(self):    
        self.logger = self.get_logger(__name__)
        
        self.params = self.get_all_input_parameters()
        self.qcm = get_qcm(self, self.params[self.InstrumentName])
        self.gate_time = self.params[self.GateTime]
        self.qcm.cmd.gate_time = self.gate_time
        self.initial_frequency = self.qcm.cmd.frequency
        self.initial_resistance = self.qcm.cmd.resistance

        self.add_details(self.initial_frequency, 'Initial frequency (Hz)')
        self.add_details(self.initial_resistance, 'Initial resistance (Ohm)')

        self.logger.info('Gate time: {} s'.format(self.params[self.GateTime]))
        
        self.ax = self.get_figure().subplots(nrows=2, ncols=1, sharex=True)

        self.resistance_plot = TimePlot(self, self.ax[1], 'Resistance', ['Resistance'])
        self.frequency_plot = TimePlot(self, self.ax[0], 'Frequency\n', ['Frequency'])
        self.frequency_plot.round_float_resolution = 9

        self.retry = 0

    def test(self):
        while self.is_running():
            try:
                new_data = self.qcm.get_data_if_both_new()
                if new_data:
                    freq = self.qcm.cmd.frequency
                    self.frequency_plot.add_data([freq], True)

                    resistance = self.qcm.cmd.resistance
                    self.resistance_plot.add_data([resistance], True)

                    # If the gate time in the input parameter frame is changed, change the gate time setting.
                    new_gate_time = self.get_input_parameter(self.GateTime)
                    if new_gate_time != self.gate_time:
                        self.gate_time = new_gate_time
                        self.qcm.cmd.gate_time = self.gate_time
                        self.logger.info('Gate time changed to {:.1f} s'.format(self.gate_time))
                    time.sleep(self.qcm.cmd.gate_time - 0.1)

                # Zoom in after the initial autoscale
                if self.frequency_plot.data_points == 2:
                    self.frequency_plot.ax.set_ylim(self.initial_frequency - 5, self.initial_frequency + 5)
                    self.resistance_plot.ax.set_ylim(self.initial_resistance - 0.5, self.initial_resistance + 0.5)

                self.retry = 0
            except Exception as e:
                self.logger.error(e)
                self.retry += 1
                if self.retry > 5:
                    break
                    
    def cleanup(self):
        pass
        




from srsgui import Instrument, SerialInterface, InstIdError
from srsgui import FindListInput
from .components import Cmd


class QCM200(Instrument):
    _IdString = 'QCM200'
    _term_char = b'\r'

    available_interfaces = [
        [
            SerialInterface,
            {
                'port': FindListInput(),
                'baud_rate': 9600,
                'hardware_flow_control': False
            }
        ],
    ]

    def __init__(self, interface_type=None, *args):
        super().__init__(interface_type, *args)

        self._over_frequency_flag = False
        self._under_frequency_flag = False
        self._comm_error_flag = False
        self._new_frequency_flag = False
        self._new_resistance_flag = False

        self.last_over_frequency_error = False
        self.last_under_frequency_error = False
        self.last_comm_error = False

        self.last_frequency = 0.0
        self.last_resistance = 0.0

        self.cmd = Cmd(self)

    def check_id(self):
        if not self.is_connected():
            return None, None, None
        reply = self.query_text('I')
        strings = reply.split('_')

        if len(strings) != 4:
            return None, None, None

        if self._IdString not in reply:
            raise InstIdError("Invalid instrument: {} not in {}"
                              .format(self._IdString, reply))

        self._id_string = reply
        self._model_name = strings[0]
        self._serial_number = strings[3]
        self._firmware_version = strings[1]
        return self._model_name, self._serial_number, self._firmware_version

    def handle_command(self, cmd_string):
        cmd = cmd_string.upper()
        reply = ''
        if cmd == 'F' or \
           cmd == 'R' or \
           cmd == 'B' or \
           cmd == 'I':
            reply = self.query_text(cmd).strip()
        else:
            reply = super().handle_command(cmd)
        return reply

    def get_data_if_both_new(self):
        status = self.cmd.status
        if status & 0x02 != 0:
            self._new_frequency_flag = True
        if status & 0x01 != 0:
            self._new_resistance_flag = True

        new_error = status & 0x1C != 0
        if new_error:
            if status & 0x4:
                self._over_frequency_flag = True
            if status & 0x8:
                self._under_frequency_flag = True
            if status & 0x10:
                self._comm_error_flag = True

        if self._new_resistance_flag and self._new_resistance_flag:
            self.last_frequency = self.cmd.frequency
            self.last_resistance = self.cmd.resistance
            self._new_resistance_flag = False
            self._new_frequency_flag = False

            self.last_over_frequency_error = True if self._over_frequency_flag else False
            self.last_under_frequency_error = True if self._under_frequency_flag else False
            self.last_comm_error = True if self._comm_error_flag else False
            last_error_state = self.last_over_frequency_error or self.last_under_frequency_error or self.last_comm_error

            self._over_frequency_flag = False
            self._under_frequency_flag = False
            self._comm_error_flag = False

            return self.last_frequency, self.last_resistance, last_error_state
        else:
            return None

    def get_status(self):
        """
        Get error status in text.
        it overrides Instrument.get_status() method.

        :rtype: str
        """

        status = self.cmd.status
        if status & 0x1C == 0:
            return 'OK'

        # Now there is an error
        over_freq_error = True if status & 0x4 else False
        under_freq_error = True if status & 0x8 else False
        comm_error = True if status & 0x10 else False

        msg = ''
        if over_freq_error:
            msg += 'Freq. Over Range Error, '
        if under_freq_error:
            msg += 'Freq. Under Range Error, '
        if comm_error:
            msg += 'Comm. Error, '
        if len(msg) > 2:
            msg = msg[:-2]
        return msg




from srsgui.inst.component import Component
from srsgui.inst.commands import IntCommand, FloatCommand, FloatGetCommand, DictCommand

from .commands import QCMGetCommand, QCMIntGetCommand, QCMFloatGetCommand


class Keys:
    AbsoluteFrequency = 'abs. F'
    RelativeFrequency = 'rel. F'
    MassDisplacement = 'mass disp.'
    AbsoluteResistance = 'abs. R'
    RelativeResistance = 'rel. R'
    GateTime = 'gate time'
    AnalogFrequencyScaleFactor = 'f scale factor'

    NewResistanceValue = 'new R'
    NewFrequencyValue = 'new F'
    FrequencyOverRange = 'F over range'
    FrequencyUnderRange = 'F under range'
    CommunicationError = 'comm. error'

    InternalTCXO = 'int. TCXO'
    External10MHz = 'ext. 10 MHz'


class Cmd(Component):
    """Component holding all remote command used in QCM200"""

    DisplayModeDict = {
        Keys.AbsoluteFrequency:  0,
        Keys.RelativeFrequency:  1,
        Keys.MassDisplacement:   2,
        Keys.AbsoluteResistance: 3,
        Keys.RelativeResistance: 4,
        Keys.GateTime:           5,
        Keys.AnalogFrequencyScaleFactor: 6
        }
    
    AnalogFrequencyScaleDict = {  # Hz/V
        200:   0,
        500:   1,
        1000:  2,
        2000:  3,
        5000:  4,
        10000: 5,
        20000: 6
        }
    
    GateTimeDict = {
        0.1:  0,
        1.0:  1,
        10.0: 2,
        }
    
    StatusBitDict = {
        Keys.NewResistanceValue:  0,
        Keys.NewFrequencyValue:   1,
        Keys.FrequencyOverRange:  2,
        Keys.FrequencyUnderRange: 3,
        Keys.CommunicationError:  4
        }
    
    TimeBaseDict = {
        Keys.InternalTCXO:  0,
        Keys.External10MHz: 1
        }
    
    id_string = QCMGetCommand('I')
    display_mode = DictCommand('D', DisplayModeDict)
    frequency_scale = DictCommand('V', AnalogFrequencyScaleDict, None, 'Hz/V')
    gate_time = DictCommand('P', GateTimeDict, None, 's')
    frequency = QCMFloatGetCommand('F', 'Hz')
    frequency_offset = FloatGetCommand('G', 'Hz', 0.0, 10000000.0, 0.01, 9)
    resistance = QCMFloatGetCommand('R', 'Ohm')
    resistance_offset = FloatGetCommand('S', 'Ohm')
    status = QCMIntGetCommand('B')
    timebase = DictCommand('T', TimeBaseDict)

    def reset_frequency_offset(self):
        self.comm.send('G')

    def reset_resistance_offset(self):
        self.comm.send('S')

    allow_run_button = [reset_frequency_offset, reset_resistance_offset]




from srsgui import Task
from srsinst.qcm.instruments.qcm200.qcm import QCM200


def get_qcm(task: Task, name=None) -> QCM200:
    inst = task.get_instrument(name)
    
    if not issubclass(inst.__class__, QCM200):
        raise TypeError('{} is not a QCM200 instance'.format(name))
    return inst




from srsgui.inst.commands import GetCommand, IntGetCommand, FloatGetCommand


class QCMGetCommand(GetCommand):
    def __init__(self, remote_command_name):
        super().__init__(remote_command_name)
        self.get_command_format = '{}'


class QCMIntGetCommand(IntGetCommand):
    def __init__(self, remote_command_name):
        super().__init__(remote_command_name)
        self.get_command_format = '{}'


class QCMFloatGetCommand(FloatGetCommand):
    def __init__(self, remote_command_name, unit=''):
        super().__init__(remote_command_name, unit)
        self.get_command_format = '{}'
